/*
 * 3D Studio to OpenGL Converter 
 * by David Farrell
 */

/*
 *  3D Studio object reader
 */

#ifdef WIN32
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include "3dsload.h"

jmp_buf EnvState;
FILE *InFile=0;
H3dsScene *Scene;
int doublesided;

/* General 3DS Reader Utility Functions */


/* Reads len number of bytes from the file into the memory pointed to by *dest */
void dread(void *dest, int len)
{
  if(fread(dest, len, 1, InFile) != 1) {
    printf("Error reading file\n");
    longjmp(EnvState, 1);
  }
}


/* Sets the file pointer to a specific position */
void dsetpos(dword pos)
{
  if(fseek(InFile, (long)pos, SEEK_SET) != 0) {
    printf("Error moving filepointer\n");
    longjmp(EnvState, 1);
  }
}


/* Retrieves the file position */
dword dgetpos(void)
{
  long pos;
  if((pos=ftell(InFile)) == -1L) {
    printf("Error getting fileposition\n");
    longjmp(EnvState, 1);
  }

  return (dword) pos;
}


/* Enlarge the scene array to hold a new mesh object */
H3dsMeshObj *EnlargeMeshArray(void) {
  void *mem;
  H3dsMeshObj *obj;

  if((mem = realloc(Scene->meshobjlist, 
		    sizeof(H3dsMeshObj)*(Scene->meshobjs+1))) == 0) {
    printf("Error reallocating mem\n");
    longjmp(EnvState, 1);
  }

  Scene->meshobjlist=(H3dsMeshObj *) mem;
  obj=&Scene->meshobjlist[Scene->meshobjs++];
  memset(obj, 0, sizeof(H3dsMeshObj));

  return obj;
}


/* Enlarge the scene array to hold a new material */
H3dsMat *EnlargeMatObjArray(void) {
  void *mem;
  H3dsMat *obj;

  if((mem = realloc(Scene->material, 
		    sizeof(H3dsMat)*(Scene->NumMaterials+1))) == 0) {
    printf("Error reallocating mem\n");
    longjmp(EnvState, 1);
  }

  Scene->material=(H3dsMat *) mem;
  obj = &Scene->material[Scene->NumMaterials++];
  memset(obj, 0, sizeof(H3dsMat));

  return obj;
}


/* Enlarges the bindings array to hold another bindings list */
H3dsMatList *EnlargeBindingsArray(H3dsMeshObj *meshobj) {
  void *mem;
  H3dsMatList *obj;

  if((mem = realloc(meshobj->bindings, 
		    sizeof(H3dsMatList)*(meshobj->NumBindings+1))) == 0) {
    printf("Error reallocating mem\n");
    longjmp(EnvState, 1);
  }

  meshobj->bindings = (H3dsMatList *)mem;
  obj = &meshobj->bindings[meshobj->NumBindings++];
  memset(obj, 0, sizeof(H3dsMatList));
    
  return obj;
}


/* Safely allocates a size amount of memory */
void *getmem(int size)
{
  void *mem;

  if((mem=malloc(size))==0) {
    printf("Error allocating mem\n");
    longjmp(EnvState, 1);
  }
  return mem;
}


/* Deletes an H3dsScene from memory */
void HFree3dsScene(H3dsScene * scene)
{
  int n,m;

  if(scene) {
    if(scene->meshobjlist) {
      for(n=scene->meshobjs-1; n>=0; n--) {
	H3dsMeshObj *mobj = &scene->meshobjlist[n];
	if(mobj->maplist)  
	  free(mobj->maplist);
	if(mobj->facelist) 
	  free(mobj->facelist);
	if(mobj->vertlist) 
	  free(mobj->vertlist);
	if (mobj->bindings) {
	  for (m=0; m<mobj->NumBindings; m++) {
	    if (mobj->bindings[m].faces) {
	      free (mobj->bindings[m].faces);
	    }
	  }
	  free (mobj->bindings);
	}
      }
      free(scene->meshobjlist);
    }
    if (scene->material) free (scene->material);
    free(scene);
  }
}


/***********************/
/* 3DS Reader Routines */
/***********************/


/* Reads in the list of vertices for the current mesh object */
void ReadVertList(H3dsMeshObj *meshobj)
{
  word nv;
  int n;

  dread(&nv, sizeof(nv));
  meshobj->NumVerts=nv;

  meshobj->vertlist=(H3dsVert *) getmem(sizeof(H3dsVert)*nv);
  for(n=0; n<nv; n++)
    dread(&meshobj->vertlist[n], sizeof(float32)*3);
}


/*
  Reads in the face list for the current mesh object.
  Each face is a set of 3 indexes into the mesh object's 
  vertex list and a set of options.
  */
void ReadFaceList(H3dsMeshObj *meshobj)
{
  word nv;
  int n;

  dread(&nv, sizeof(nv));
  meshobj->NumFaces=nv;

  meshobj->facelist=(H3dsFace *) getmem(sizeof(H3dsFace)*nv);
  for(n=0; n<nv; n++)
    dread(&meshobj->facelist[n], sizeof(word)*4);
}


/* Reads in the mapping list for the current mesh object */
void ReadMapList(H3dsMeshObj *meshobj)
{
  word nv;
  int n;

  dread(&nv, sizeof(nv));
  meshobj->NumMaps=nv;

  meshobj->maplist=(H3dsMap *) getmem(sizeof(H3dsMap)*nv);
  for(n=0; n<nv; n++)
    dread(&meshobj->maplist[n], sizeof(float32)*2);
}


/* 
  Reads in the translation matrix for the current mesh object.
  I'm not sure when this is used in a 3DS model.
  */
void ReadTraMatrix(H3dsMeshObj *meshobj)
{
  dread(&meshobj->TraMatrix, sizeof(float32)*12);
  meshobj->matrix=1;
}


/*
  Assigns the material in this chunk to the faces specified. 
  Each material has a list of indexes into the face array to
  mark which faces use which material.
  */
void ReadFaceMaterial(H3dsMeshObj *meshobj)
{
  word nf;
  word face;
  char name[50];
  H3dsMatList *binding;
  int n;

  binding = EnlargeBindingsArray(meshobj);

  /* The material name is first */
  n=0;
  do {
    dread(&name[n++], 1);
  } while(name[n-1]!='\0' && n<sizeof(name));
  name[n-1]='\0';
  strcpy(binding->name, name);

  /* Then the number of faces */
  dread(&nf, sizeof(nf));
  binding->NumFaces = (int)nf;
  binding->faces = (word *)(malloc(sizeof(face)*nf));

  /* Bind the faces to the material */
  for (n=0; n<nf; n++) {
    dread(&face, sizeof(face));
    binding->faces[n] = face;
  }
} 
  
/*
  Eats a triangle mesh, which is a combination of XYZ vertices,
  faces, and bindings of material to faces. 
  */
void ReadTriMeshBlocks(dword p, char *name)
{
  word id;
  dword len, pc;
  H3dsMeshObj *meshobj;

  meshobj = EnlargeMeshArray();

  strcpy(meshobj->name, name);
  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((int)id) {
      case CHUNK_VERTLIST: ReadVertList (meshobj); break;
      case CHUNK_FACELIST: ReadFaceList (meshobj); break;
      case CHUNK_MAPLIST:  ReadMapList  (meshobj); break;
      case CHUNK_TRMATRIX: ReadTraMatrix(meshobj); break;
      case CHUNK_FACEMAT:  ReadFaceMaterial(meshobj); break;
	/* case CHUNK_SMOOLIST: */
      default: dsetpos(pc+len);
    }
  }
}


/*
  Reads in object blocks, which are cameras, meshes, or lights;
  this code only supports meshes and ignores the rest
  */
void ReadObjBlocks(dword p)
{
  word id;
  dword len, pc;
  char name[16];

  /* The object name is always the first item */
  int n=0;
  do {
    dread(&name[n++], 1);
  } while(name[n-1]!='\0' && n<sizeof(name));
  name[n-1]='\0';

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((int)id) {
      case CHUNK_TRIMESH: ReadTriMeshBlocks(pc+len, name); break;
	/*
	  case CHUNK_LIGHT:
	  case CHUNK_CAMERA:
	  */
      default: dsetpos(pc+len);
    }
  }
}


/* Reads in the name of the material */
void ReadMatChunk(dword p, H3dsMat *mat) {
  int n;
  char name[16];
  
  n = 0;
  do {
    dread(&name[n++], 1);
  } while(name[n-1]!='\0' && n<sizeof(name));
  name[n-1]='\0';

  strcpy(mat->name, name);
}


/* Reads in an RGB FP triplet */
void ReadRGBFloat(dword p, H3dsRGBColor *rgb) {
  dread(&rgb->red, sizeof(float32));
  dread(&rgb->green, sizeof(float32));
  dread(&rgb->blue, sizeof(float32));

}


/* Reads in an RGB byte triplet */
void ReadRGBByte(dword p, H3dsRGBColor *rgb) {
  unsigned char red, green, blue;

  dread(&red, sizeof(red));
  dread(&green, sizeof(green));
  dread(&blue, sizeof(blue));
  rgb->red = (float)(red)/256.0;
  rgb->green = (float)(green)/256.0;
  rgb->blue = (float)(blue)/256.0;
}


/* Reads in a color assignment */
void ReadShadingChunk(dword p, H3dsRGBColor *rgb) {
  word id;
  dword len, pc;

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((unsigned int)id) {
      case CHUNK_RGB1: ReadRGBFloat(pc+len, rgb); break;
      case CHUNK_RGB2: ReadRGBByte (pc+len, rgb); break;
      default: dsetpos(pc+len);
    }
  }
}


void ReadTextureNameChunk(dword p, H3dsTexture *texture) {
  int n;
  char name[16];
  word id;
  dword len, pc;

  n = 0;
  do {
    dread(&name[n++], 1);
  } while(name[n-1]!='\0' && n<sizeof(name));
  name[n-1]='\0';
  
  strcpy(texture->texturename, name);
  printf(" Texture: %s\n", name);
}

void ReadTextureChunk(dword p, H3dsTexture *texture) {
  unsigned short id;
  dword len, pc;

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((unsigned int)id) {
      case CHUNK_MAPFILENAME: ReadTextureNameChunk(pc+len, texture); break;
      default: dsetpos(pc+len);
    }
  }
}  

/*
  Fills out the current material with the information contained 
  in these chunks
  */
void ReadMaterialBlocks(dword p)
{
  unsigned short id;
  dword len, pc;

  H3dsMat *materialobj = EnlargeMatObjArray();

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));

    switch((unsigned int)id) {
    case CHUNK_MATNAME: 
      ReadMatChunk(pc+len, materialobj); 
      break;
    case CHUNK_AMBIENT: 
      ReadShadingChunk(pc+len, &materialobj->ambient);
      break;
    case CHUNK_DIFFUSE: 
      ReadShadingChunk(pc+len, &materialobj->diffuse);
      break;
    case CHUNK_SPECULAR:
      ReadShadingChunk(pc+len, &materialobj->specular);
      break;
    case CHUNK_TEXTURE:
      ReadTextureChunk(pc+len, &materialobj->texture);
    default: 
      dsetpos(pc+len);
      break;
    }
  }
}


/* Reads in mesh objects and materials */
void ReadObjMeshBlocks(dword p)
{
  unsigned short id;
  dword len, pc;

  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((unsigned int)id) {
      case CHUNK_MATERIAL: ReadMaterialBlocks(pc+len); break;
      case CHUNK_OBJBLOCK: ReadObjBlocks(pc+len); break;
	
      default: dsetpos(pc+len);
    }
  }
}


/*
  Reads in objects; this would be the start of where keyframer support would
  be added
  */
void ReadMainBlocks(dword p)
{
  word id;
  dword len, pc;
  
  while((pc=dgetpos()) < p) {
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    switch((int)id) {
      case CHUNK_OBJMESH: ReadObjMeshBlocks(pc+len); break;
	/* case CHUNK_KEYFRAMER: */
      default: dsetpos(pc+len);
    }
  }
}


/* Reads in a H3dsScene from the disk */
H3dsScene *HRead3dsScene(void *ptr, int what, dword size) 
{
  int retval = setjmp(EnvState);
  
  if(ptr==0)
    return 0;
  
  if(what==0)
    InFile=(FILE *) ptr;
  else
    return 0;
  
  /* Initialize the global var Scene & fill with zeroes */
  if ((Scene=(H3dsScene *) malloc(sizeof(H3dsScene)))==0)
    return 0;
  memset(Scene, 0, sizeof(H3dsScene));
  Scene->material = 0;
  Scene->NumMaterials = 0;

  if(retval==0) {
    /* Return address set, start loading 3DS data. */
    word id;
    dword len, pc;

    pc=dgetpos();
    dread(&id, sizeof(id));
    dread(&len, sizeof(len));
    if((int)id!=CHUNK_MAIN) {
      HFree3dsScene(Scene);
      printf("Not 3ds data\n");
      return 0;
    }
    ReadMainBlocks(pc+len);
  }
  else {
    /* There was an error, free the allocated mem and return NULL. */
    HFree3dsScene(Scene);
    return 0;
  }
  
  return Scene;
}


/* Feed our internal data to the outgl.c functions */
void convert_to_outgl() {
  int x,y,z;
  int current;
  int MaterialNum;
  void *ptr;
  RGB ambient, diffuse, specular;
  VERTEX v;
  TRIANGLE tri;
  char *name;

  /* Start with the material properties */
  for (x=0; x<Scene->NumMaterials; x++) {
    H3dsMat *mat = &Scene->material[x];
    ambient.red = mat->ambient.red;
    ambient.green = mat->ambient.green;
    ambient.blue = mat->ambient.blue;
    ambient.alpha = 1.0;
    diffuse.red = mat->diffuse.red;
    diffuse.green = mat->diffuse.green;
    diffuse.blue = mat->diffuse.blue;
    diffuse.alpha = 1.0;
    specular.red = mat->specular.red;
    specular.green = mat->specular.green;
    specular.blue = mat->specular.blue;
    specular.alpha = 1.0;
    OutGLAddMaterial(ambient, diffuse, specular, mat->name);
  }

  /* Then do the vertices and polygons */
  /* Order is important! */
  current = 0;
  for (x=0; x<Scene->meshobjs; x++) {
    H3dsMeshObj *meshobj = &Scene->meshobjlist[x];

    /* Feed vertices to outgl */
    for (y=0; y<meshobj->NumVerts; y++) {
      v.x = meshobj->vertlist[y].x;
      v.y = meshobj->vertlist[y].y;
      v.z = meshobj->vertlist[y].z;
      OutGLAddVertex(v);
    }

    for (y=0; y<meshobj->NumBindings; y++) {
      name = meshobj->bindings[y].name;
      for (z=0; z<meshobj->bindings[y].NumFaces; z++) {
	tri.p0 = meshobj->facelist[meshobj->bindings[y].faces[z]].p0 + current;
	tri.p1 = meshobj->facelist[meshobj->bindings[y].faces[z]].p1 + current;
	tri.p2 = meshobj->facelist[meshobj->bindings[y].faces[z]].p2 + current;
	OutGLAddPolygon(tri, name);
	/* I can't figure out how to extract doublesided vs. singlesided information */
	/* from .3DS files-- if anyone knows how, please email me */
	if (doublesided) {
	  tri.p0 = meshobj->facelist[meshobj->bindings[y].faces[z]].p0 + current;
	  tri.p1 = meshobj->facelist[meshobj->bindings[y].faces[z]].p2 + current;
	  tri.p2 = meshobj->facelist[meshobj->bindings[y].faces[z]].p1 + current;
	  OutGLAddPolygon(tri, name);
	}
      }
    }
    current += meshobj->NumVerts;
  }

  /* Wrap everything up in the outgl object */
  OutGLFinishModel();
}      


/*
  Function to read in the .3DS file *filename and assign its contents to
  the polygon and object classes. This kicks the whole 3DS parsing off.
*/
int parse3ds (char *filename, int double_sided,int (*callback)(void)) 
{
  FILE *inf;
  H3dsScene *scene;
  long size;
  int retval;
  
  doublesided = double_sided;

  /* Open & prepare input file */
  if(!(inf = fopen(filename, "rb"))) {
    printf("Failed to open %s\n", filename);
    return 1;
  }

  if(fseek(inf, 0, SEEK_END)) {
    printf("Error seeking %s\n", filename);
    fclose(inf);
    return 1;
  }
  
  if((size=ftell(inf)) == -1L) {
    printf("Error sizing %s\n", filename);
    fclose(inf);
    return 1;
  }
  rewind(inf);
  
  /* Read in the file and place it in *scene */
  if((scene = HRead3dsScene(inf, 0, size)) == 0) {
    printf("Failed to parse %s\n", filename);
    fclose(inf);
    return 1;
  }
  
  fclose(inf);

	retval=callback();

  HFree3dsScene(scene);
  
  printf("\n");
  return retval;
}
