#define		STRICT
#include	<windows.h>
#include	<libv.h>

/*
 *	IPC API to use for communication between threads, processes and
 *	machines across the local net / Net.
 *	We use client-server ideology: IPC channel first opened at the
 *	server side by ipcServerEndpoint(), and than clients can connect
 *	to channel by ipcClientEndpoint().
 *	We won't use callbacks because they will be called in context of
 *	ipc thread, not caller's context; this may cause problems - libV
 *	isolate threads from each other. Instead, we accumulate received
 *	data and provide ipcRecvCount() which returns amount of it.
 *	Notes.
 *	Only local IPC currently implemented.
 */

struct ipc_channel_t, *ipc_handle_t;

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*	ipcServerEndpoint(): creates IPC channel and begins listening for
	 *	client connections.
	 *	Input:
	 *		channel_name	- name of server.
	 *	Output:
	 *		IPC handle or NULL on error.
	 */
	ipc_handle_t	V_GLOBAL	ipcServerEndpoint(char *channel_name);
	/*	ipcClientEndpoint(): opens IPC channel.
	 *	Input:
	 *		channel_name	- name of server.
	 *	Output:
	 *		IPC handle or NULL if no server.
	 */
	ipc_handle_t	V_GLOBAL	ipcClientEndpoint(char *channel_name);
	/*	ipcCloseEndpoint():
	 *		- for server side: disconnects all clients and destroys
	 *		channel;
	 *		- for client side: disconnects client and destroys
	 *		channel.
	 *	Input:
	 *		ipch - IPC handle.
	 *	Output:
	 *		None; IPC handle becomes invalid.
	 */
	void			V_GLOBAL	ipcCloseEndpoint(ipc_handle_t ipch);
	/*	ipcSendCount(): returns number of bytes queued to send.
	 *	Input:
	 *		ipch - IPC handle.
	 *	Output:
	 *		Number of bytes in output queue.
	 */
	unsigned		V_GLOBAL	ipcSendCount(ipc_handle_t ipch);
	/*	ipcSend(): queues data to send and returns immediately.
	 *	Input:
	 *		ipch - IPC handle;
	 *		src - buffer to send;
	 *		num_bytes - number of bytes to send.
	 *	Output:
	 *		None (caller should use ipcSendCount() to monitor transmission progress).
	 */
	void			V_GLOBAL	ipcSend(ipc_handle_t ipch,void *src,unsigned num_bytes);
	/*	ipcRecvCount(): returns number received bytes.
	 *	Input:
	 *		IPC handle.
	 *	Output:
	 *		Number of bytes in input queue.
	 */
	unsigned		V_GLOBAL	ipcRecvCount(ipc_handle_t ipch);
	/*	ipcRecv(): copies num_bytes from input queue to dst.
	 *	Input:
	 *		ipch - IPC handle;
	 *		dst - buffer to hold received data;
	 *		num_bytes - number of bytes to receive.
	 *	Output:
	 *		true - if at least num_bytes was in input queue;
	 *		false - otherwise.
	 *	If function fails, NO bytes stored in 'dst' (and, therefore, size
	 *	of input queue unchanged).
	 */
	vboolean		V_GLOBAL	ipcRecv(ipc_handle_t ipch,void *dst,unsigned num_bytes);
#ifdef	__cplusplus
}
#endif	/* __cplusplus */


/*	IMPLEMENTATION.
 */

/* The only type of IPC currently implemented */
#define		IPC_TYPE_LOCAL			1

/* Maximal length of IPC name */
#define		MAX_IPC_NAME			_MAX_PATH

/*	======================================================================
 *	Local IPC implemented via shared memory.
 *	Both NT and UNIX have API to create shared memory mappings. There is
 *	no such API under DOS; but DOS also have no multitasking, so we can
 *	simple malloc() block and use it.
 */

struct	shm_obj_s
{
#if		defined(__WIN32__)
	/*	NT stores names of shared mappings in kernel, so we only must
	 *	keep map handle and base.
	 */
	HANDLE	map;
	void_p	map_base;
#elif	defined(__UNIX__)
	/*	UNIX stores integer identifiers of mappings, so here we must
	 *	keep both map name, map ID and map base.
	 */
	int		map_id;
	char	map_name[MAX_IPC_NAME];
	void_p	map_base;
#elif	defined(__DOS32__)
	/*	Someone still using this "operating system" ?
	 *	No handle present; we keep only name and base address.
	 */
	char	map_name[MAX_IPC_NAME];
	void_p	map_base;
#endif
}	shm_obj_t, *shm_obj_p;

#define		IPC_MAP_CHUNK			(64*1024)

#if		defined(__DOS32__) || defined(__UNIX)
/*
 *	Under DOS and UNIX we use static storage for all shm objects.
 */
static	vp_stack_t	named_maps = {0, 0, NULL};

#endif	/* defined(__DOS32__) || defined(__UNIX) */

shm_obj_p	V_LOCAL
CreateNamedSharedMemory(char *map_name,vboolean open_existing)
{
	shm_obj_p shm;
#if		defined(__WIN32__)
	shm=(shm_obj_t*)malloc(sizeof(shm_obj_t));
	v_zero(shm,sizeof(shm_obj_t));
	/* Create mapping */
	if(open_existing)
	{
		shm->map=OpenFileMapping(FILE_MAP_READ|FILE_MAP_WRITE,FALSE,
			map_name);
		if(shm->map == NULL)
		{
#ifdef	CHECKED_BUILD
			conlPrintf("CreateNamedSharedMemory(): OpenFileMapping() failed "
				"(error code %d).\n", GetLastError());
#endif	/* CHECKED_BUILD */
			free(shm);
			return NULL;
		}
	}
	else	/* open_existing == 0 */
	{
		shm->map=CreateFileMapping((HANDLE)0xFFFFFFFF,NULL,PAGE_READWRITE | 
			SEC_COMMIT | SEC_NOCACHE, 0, IPC_MAP_CHUNK,map_name);
		if(shm->map == NULL)
		{
#ifdef	CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"CreateNamedSharedMemory(): CreateFileMapping() failed "
				"(error code %d).\n", GetLastError());
#endif	/* CHECKED_BUILD */
			free(shm);
			return NULL;
		}
		if(GetLastError() == ERROR_ALREADY_EXISTS)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,"CreateNamedSharedMemory(): object '%s' "
				"already exists.\n",map_name);
#endif		/* CHECKED_BUILD */
			CloseHandle(shm->map);
			free(shm);
			return NULL;
		}
	}
	/* Get map base address */
	shm->map_base=MapViewOfFile(shm->map,FILE_MAP_READ|FILE_MAP_WRITE,0,0,
		IPC_MAP_CHUNK);
	if(shm->map_base == NULL)
	{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"CreateNamedSharedMemory(): MapViewOfFile() failed "
				"(error code %d).\n",GetLastError());
#endif		/* CHECKED_BUILD */
		CloseHandle(shm->map);
		free(shm);
		return NULL;
	}
	/* We are succeded ! */
	return shm;

#elif	defined(__UNIX__)
	/*	Under UNIX (and DOS) we must manually track namespace of
	 *	IPC channels. It's bit tricky under Unix: we must map
	 *	from name to integer id, and for that dedicated 'name
	 *	server' is required. We start server at first failed
	 *	call to UnixConnectToIpcNameServer().
	 */
	
	if(!UnixConnectToIpcNameServer())
	{
		if(!UnixStartIpcNameServer())
		{
			return NULL;
		}
	}
	map_id=UnixMapIpcNameToIpcId(map_name);
	
	/*
	shm->id=shm_get();
	shm->map_base=shm_at();
	*/
	
	return NULL;

#elif	defined(__DOS32__)

	shm_obj_p ptr;
	int slot;
	for(slot=0;slot<vpsCount(&named_maps);slot++)
	{
		ptr=(shm_obj_t*)vpsGet(&names_maps,slot);
		if(!strncmp(map_name,ptr->map_name,MAX_MAP_NAME))
			break;
	}
	if(open_existing)
	{
		/*	For client: named memory with name 'map_name' MUST be
		 *	already created.
		 */
		if(slot==vpsCount(&named_maps))
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"CreateNamedSharedMemory(): map '%s' not found.\n",
				map_name);
#endif		/* CHECKED_BUILD */
			return NULL;
		}
		/* Return previously allocated object */
		return ptr;
	}
	else
	{
		/*	For server: named memory with name 'map_name' MUST NOT be
		 *	already created.
		 */
		if(slot<vpsCount(&named_maps))
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"CreateNamedSharedMemory(): map '%s' already exists.\n",
				map_name);
#endif		/* CHECKED_BUILD */
			return NULL;
		}
		/* Allocate new object */
		shm=(shm_obj_t*)malloc(sizeof(shm_obj_t));
		v_zero(shm,sizeof(shm_obj_t));
		strCopy(shm->map_name,map_name,MAX_MAP_NAME);
		chm->map_base=malloc(IPC_MAP_CHUNK);
		if(chm->map_base==NULL)
		{
#ifdef		CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"CreateNamedSharedMemory(): malloc(%d) failed.\n",
				IPC_MAP_CHUNK);
#endif		/* CHECKED_BUILD */
			free(shm);
			return NULL;
		}
		/* Store object in 'named_maps' */
		for(slot=0;slot<vpsCount(&named_maps);slot++)
			if(vpsGet(&names_maps,slot)==NULL)
				break;
		if(slot==vpsCount(&named_maps))
			vpsPush(&named_maps);
		else
			vpsGetStorage(&named_maps)[slot]=shm;
		return shm;
	}

#endif	/* defined(__DOS32__) */

	/* Unknown platform ! */
	
	return NULL;
}

void	V_LOCAL
DestroyNamedSharedMemory(shm_obj_t *shm)
{
	if(shm==NULL)
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_WARNING,
			"DestroyNamedSharedMemory([NULL]): incorrect parameter.\n",
			IPC_MAP_CHUNK);
#endif	/* CHECKED_BUILD */
		return;
	}
#if		defined(__WIN32__)
	/* Under NT we simply unmaps and releases handle */
	if(shm->map_base)
		UnmapViewOfFile(shm->map_base);
	if(shm->map)
		CloseHandle(shm->map);
	free(*shm);
#elif	defined(__UNIX__)
	/*	We must stop nameserver if there is no more entries
	 *	in map of them.
	 */
	if(!UnixGetIpcNameserverMapCount())
		UnixShutdownIpcNameserver();
	
#elif	defined(__DOS32__)
	/* There some disadvantages of pure C ... */
	{
		int slot;
		/* Remove reference to object from 'named_maps' */
		for(slot=0;slot<vpsCount(&named_maps);slot++)
			if(vpsGet(&names_maps,slot)==NULL)
				break;
		if(slot==vpsCount(&named_maps))
		{
#ifdef		CHECKED_BUILD
			conlPuts(LOG_WARNING,
				"DestroyNamedSharedMemory('%s'): map not exists.\n",
				shm->map_name);
#endif		/* CHECKED_BUILD */
			/* 'shm' isn't correct map object; therefore, we don't 
			 *	attempt to free() it. */
			return;
		}
		else
			vpsGetStorage(&named_maps)[slot]=NULL;
		free(shm->map_base);
		free(shm);
	}

#endif

	return;
}

/*	======================================================================
 *	Local IPC implementation.
 *	Main idea.
 *		1. Create memory which can be shared by many processes.
 *		2. Create access restricting objects (semaphores).
 *		3. Create signaling objects (events).
 *		4. Start separate thread.
 *		5a. For server thread:
 *			1. Read ipc_cmd dword from beginning of map.
 *			2. If there is data to read - copy it from map to in_queue.
 *			2a. If nothing to read, check out_queue.
 *			3. Wait for signal and than goto 1.
 */




