/*<html><body bgcolor=#B4C4B0><font color=#000000><strong><pre>*/
#ifndef	V_IPC_H
#define	V_IPC_H

/*
 *	IPC API to use for communication between threads, processes and
 *	machines across the local net / Net.
 *	We use client-server ideology: IPC channel first opened at the
 *	server side by ipcServerEndpoint(), and than clients can connect
 *	to channel by ipcClientEndpoint().
 *	We won't use recv- and send- callbacks because they will be called
 *	in the context of ipc thread, not caller's context; this may cause
 *	problems. Instead, we:
 *		a) accumulate received data and provide ipcRecvCount() which
 *		returns amount of it;
 *		b) provide ipcWait() which should be used to wait for received
 *		data.
 *	Notes.
 *	Only thread - level IPC currently implemented.
 *	TODO: Implement machine- and net-level IPC.
 */

typedef	u32_t	ipc_handle_t;

/*	Infinite timeout */
#ifndef	INFINITE
#define	INFINITE	0xFFFFFFFF
#endif	/* INFINITE */

#ifdef	__cplusplus
extern "C" {
#endif	/* __cplusplus */
	/*	This must be called before any IPC operations */
	void		V_GLOBAL	ipcInit(void);
	/*	This must be called after all IPC operations */
	void		V_GLOBAL	ipcShutdown(void);
	/*	ipcServerEndpoint(): creates IPC channel and begins listening for
	 *	client connections.
	 *	Input:
	 *		channel_name	- name of server.
	 *	Output:
	 *		IPC handle or NULL on error.
	 */
	ipc_handle_t	V_GLOBAL	ipcServerEndpoint(char *channel_name);
	/*	ipcClientEndpoint(): opens IPC channel.
	 *	Input:
	 *		channel_name	- name of server.
	 *	Output:
	 *		IPC handle or NULL if no server.
	 */
	ipc_handle_t	V_GLOBAL	ipcClientEndpoint(char *channel_name);
	/*	ipcClose():
	 *		- for server side: disconnects all clients and destroys
	 *		channel.
	 *		- for client side: disconnects client.
	 *	Input:
	 *		ipch - IPC handle.
	 *	Output:
	 *		None; IPC handle becomes invalid.
	 */
	void			V_GLOBAL	ipcClose(ipc_handle_t ipch);
	/*	ipcSendCount(): returns number of bytes queued to send.
	 *	Input:
	 *		ipch - IPC handle.
	 *	Output:
	 *		Number of bytes in output queue.
	 */
	u32_t			V_GLOBAL	ipcSendCount(ipc_handle_t ipch);
	/*	ipcSendTo():
	 *		SERVER-SIDE: queues data to send to specified client and returns
	 *			immediately.
	 *		CLIENT-SIDE: can't be used.
	 *	Input:
	 *		ipch - IPC handle;
	 *		cl_id - ID of client;
	 *		src - buffer to send;
	 *		num_bytes - number of bytes to send.
	 *	Output:
	 *		None (caller should use ipcSendCount() to monitor transmission
	 *		progress).
	 */
	void			V_GLOBAL	ipcSendTo(ipc_handle_t ipch,u32_t cl_id,
		void *src,u32_t num_bytes);
	/*	ipcSend():
	 *		CLIENT-SIDE: queues data to send to server.
	 *		SERVER-SIDE: queues data to send to all clients.
	 *	Input:
	 *		ipch - IPC handle;
	 *		src - buffer to send;
	 *		num_bytes - number of bytes to send.
	 *	Output:
	 *		None (caller should use ipcSendCount() to monitor transmission
	 *		progress).
	 */
	void			V_GLOBAL	ipcSend(ipc_handle_t ipch,void *src,
		u32_t num_bytes);
	/*	ipcRecvCount(): returns number of received bytes.
	 *	Input:
	 *		IPC handle.
	 *	Output:
	 *		Number of bytes in input queue.
	 */
	u32_t			V_GLOBAL	ipcRecvCount(ipc_handle_t ipch);
	/*	ipcRecv(): copies num_bytes from input queue to dst.
	 *	Input:
	 *		ipch - IPC handle;
	 *		dst - buffer to hold received data;
	 *		num_bytes - number of bytes to receive.
	 *	Output:
	 *		true - if at least num_bytes was in input queue;
	 *		false - otherwise.
	 *	If function fails, NO bytes stored in 'dst' (and, therefore, size
	 *	of input queue unchanged).
	 */
	vboolean		V_GLOBAL	ipcRecv(ipc_handle_t ipch,void *dst,
		u32_t num_bytes);
	/*	ipcGets(): gets string from ipc channel.
	 *	Input:
	 *		ipch - IPC handle;
	 *		dst - buffer to hold string;
	 *		dst_len - length of buffer.
	 *	Output:
	 *		Number of characters stored in dst (not more than dst_len-1).
	 *	Notes.
	 *		ipcGets() returns immediately if no data to receive.
	 */
	int				V_GLOBAL	ipcGets(ipc_handle_t ipch,char *dst,
		int dst_len);
	/*
	 *	Connect and disconnect callbacks will always be called in context
	 *	of caller (IPC channel creator) thread.
	 */
	void			V_GLOBAL	ipcConnectCallback(ipc_handle_t ipch,
		void (V_CALL *connect_callback)(ipc_handle_t ipch,u32_t cl_id));
	void			V_GLOBAL	ipcDisconnectCallback(ipc_handle_t ipch,
		void (V_CALL *disconnect_callback)(ipc_handle_t ipch,u32_t cl_id));
	/*
	 *	ipcWait(): blocks caller's thread if no data in input queue.
	 *	Input:
	 *		ipch - handle of channel (can be both server and client);
	 *		timeout - timeout in milliseconds.
	 *	Output:
	 *		true - there is data to ipcRecv();
	 *		false - timeout, no data.
	 */
	vboolean		V_GLOBAL	ipcWait(ipc_handle_t ipch,u32_t timeout);
	/*
	 *	Returns name of server for given slot (may be client's slot).
	 */
	vboolean		V_GLOBAL	ipcGetServerName(ipc_handle_t ipch,char *dest);
	/*
	 *	Returns client's ID for given client's slot.
	 */
	vboolean		V_GLOBAL	ipcGetClientId(ipc_handle_t ipch,u32_t *id);
	/*
	 *	Server side: clears input and/or all output queues.
	 *	Client side: clears output queue (input queue can only be cleared
	 *		from server side).
	 */
	vboolean		V_GLOBAL	ipcPurge(ipc_handle_t ipch,
		vboolean purge_input,vboolean purge_output);
#ifdef	__cplusplus
}
#endif	/* __cplusplus */

#endif	/* V_IPC_H */
/*</pre></strong></body></html>*/
/*EOF*/
